

import os
import re
import time
import numpy as np
import sbmlxdf

from modelpruner.models.fba_model import FbaModel
from modelpruner.core.protected_parts import ProtectedParts


class ModelPruner:

    def __init__(self, sbml_fname, protected_parts, reduced_fname=None):

        if reduced_fname is None:
            self.reduced_fname = re.sub('.xml', '_red.xml', sbml_fname)

        if os.path.exists(sbml_fname):
            self.sbml_model = sbmlxdf.Model(sbml_fname)
            print(f'Protected data loaded: {sbml_fname} '
                  f'(last modified: {time.ctime(os.path.getmtime(sbml_fname))})')
        else:
            print(f'{sbml_fname} not found!')
            raise FileNotFoundError

        self.tolerance = 1e-8   # later accessible by set_params() ?

        self.fba_model = FbaModel(self.sbml_model)
        self.nrp = ProtectedParts(protected_parts)

        self.base_obj_dir = self.fba_model.obj_dir
        self.base_flux_bounds = self.fba_model.flux_bounds.copy()
        self.base_coefs = self.fba_model.obj_coefs.copy()
        self.protected_sids = set(self.nrp.initial_protected_sids)
        self.protected_rids = self.get_total_protected_rids(set(self.nrp.initial_protected_rids))
        self.essential_rids = self.protected_rids
        self.free_rids = set(self.fba_model.rids).difference(self.essential_rids)
        self.set_function_optimum()
        self.check_protected_parts()

        # overwrite bounds on fba
        self.fba_model.update_flux_bounds(self.nrp.overwrite_bounds)

    def check_protected_parts(self):

        extra_p_rids = self.protected_rids.difference(set(self.fba_model.rids))
        extra_p_sids = self.protected_sids.difference(set(self.fba_model.sids))
        pf_rids = set()
        for pf in self.nrp.protected_functions.values():
            pf_rids = pf_rids.union({rid for rid in pf.objective})
            pf_rids = pf_rids.union({rid for rid in pf.overwrite_bounds})
            if pf.fba_success is False:
                print(f'protected function {pf.obj_id} has unsuccessful FBA '
                      f' with optimal value {pf.optimal}')

        extra_pf_rids = pf_rids.difference(set(self.fba_model.rids))
        if len(extra_p_rids) > 0:
            print('protected reactions that do not exist in model:', extra_p_rids)
        if len(extra_p_sids) > 0:
            print('protected metabolites that do not exist in model:', extra_p_sids)
        if len(extra_pf_rids) > 0:
            print('reactions in protected function that do not exist in model:', extra_pf_rids)

        zero_flux_rids = self.get_blocked_reactions()
        blocked_p_rids = self.protected_rids.intersection(zero_flux_rids)
        blocked_pf_rids = pf_rids.intersection(zero_flux_rids)
        if len(blocked_p_rids) > 0:
            print('protected reactions that are blocked (no flux):', blocked_p_rids)
        if len(blocked_pf_rids) > 0:
            print('reactions in protected function that are blocked (no flux):', blocked_pf_rids)

    def get_blocked_reactions(self):
        # identify blocked reactions
        # check that none of the blocked reactions are protected
        # actually remove blocked reactions, even they are included in

        fba_solution = self.fba_model.fba_optimize()
        zero_flux_mask = np.abs(fba_solution['x']) < self.tolerance
        zero_flux_rids = self.fba_model.rids[zero_flux_mask]

        fva_solution = self.fba_model.fva_optimize(zero_flux_rids, fract_of_optimum=0.0)
        flux_ranges = np.vstack((fva_solution['min'], fva_solution['max'])).T
        zero_flux_mask = np.all(abs(flux_ranges) < self.tolerance, axis=1)
        zero_flux_rids = set(fva_solution['rids'][zero_flux_mask])
        return zero_flux_rids

    def prune(self):

        self.remove_parallel_reactions()

    def reaction_types_fva(self):

        free_rids = list(set(self.fba_model.rids).difference(self.protected_rids))

        flux_min_pfs = np.zeros((len(self.nrp.protected_functions), len(free_rids)))
        flux_max_pfs = np.zeros((len(self.nrp.protected_functions), len(free_rids)))

        for idx, pf in enumerate(self.nrp.protected_functions.values()):
            self.set_temp_func_conditions(pf)
            res = self.fba_model.fva_optimize(free_rids, fract_of_optimum=0.0)
            self.restore_base_conditions()
            flux_min_pfs[idx] = res['min']
            flux_max_pfs[idx] = res['max']

        flux_min_pfs[np.abs(flux_min_pfs) < self.tolerance] = 0.0
        flux_max_pfs[np.abs(flux_max_pfs) < self.tolerance] = 0.0
        flux_min = np.min(flux_min_pfs, axis=0)
        flux_max = np.max(flux_max_pfs, axis=0)

        mask_b = np.all(np.vstack((flux_min == 0, flux_max == 0)), axis=0)
        blocked_rids = set(np.array(free_rids)[mask_b])

        mask_e1 = np.all(np.vstack((flux_min < 0, flux_max < 0)), axis=0)
        mask_e2 = np.all(np.vstack((flux_min > 0, flux_max > 0)), axis=0)
        mask_e = np.any(np.vstack((mask_e1, mask_e2)), axis=0)
        essential_rids = set(np.array(free_rids)[mask_e])

        mask_c = np.all(np.vstack((np.logical_not(mask_b), np.logical_not(mask_e))), axis=0)
        flux_span = flux_max - flux_min
        ind = np.argsort(flux_span)

        candidate_rids_sorted = list(np.array(free_rids)[ind][np.array(mask_c)[ind]])
        return {'blocked_rids': blocked_rids, 'essential_rids': essential_rids,
                'candidate_rids_sorted': candidate_rids_sorted}

    def remove_parallel_reactions(self):

        parallel_rid_groups = self.fba_model.get_identical_columns()

        drop_rids = set()
        for group_rids in parallel_rid_groups:
            protected = set()
            reversible = None
            for rid in group_rids:
                if rid in self.protected_rids:
                    protected.add(rid)
                if self.fba_model.reversible[self.fba_model.rid2idx[rid]]:
                    reversible = rid
            if len(protected) > 0:
                drop_group_rids = set(group_rids) - protected
            elif reversible is not None:
                drop_group_rids = set(group_rids) - {reversible}
            else:
                drop_group_rids = set(group_rids[1:])
            drop_rids = drop_rids.union(drop_group_rids)

        self.fba_model.remove_reactions(drop_rids, self.protected_sids)
        print(f'{len(drop_rids)} parallel reactions dropped:', drop_rids)

    def set_function_optimum(self):
        """using FBA to set optimal value for protected functions"""
        for obj_id in self.nrp.protected_functions:
            pf = self.nrp.protected_functions[obj_id]
            self.set_temp_func_conditions(pf)

            res = self.fba_model.fba_optimize(short_result=True)

            pf.optimal = res['fun']
            pf.fba_success = res['success']
            if hasattr(pf, 'target_frac'):
                if pf.obj_dir == 'minimize':
                    pf.set_target_ub(pf.target_frac * pf.optimal)
                else:
                    pf.set_target_lb(pf.target_frac * pf.optimal)

            self.restore_base_conditions()

    def get_total_protected_rids(self, protected_rids):
        # identify reactions that need protection due to protected metabolites
        #  over and above the protected rids provided in protected parts
        sid_mask = np.array([True if sid in self.protected_sids else False
                             for sid in self.fba_model.sids])

        psf_mask = np.any(self.fba_model.s_mat[sid_mask] != 0, axis=0)
        protected_sid_rids = set(self.fba_model.rids[psf_mask])

        # identify rids in protected functions and ensure they are protected
        pf_rids = set()
        for pf in self.nrp.protected_functions.values():
            pf_rids = pf_rids.union({rid for rid in pf.objective})
            pf_rids = pf_rids.union({rid for rid in pf.overwrite_bounds})

        total_protect_rids = protected_rids.union(protected_sid_rids).union(pf_rids)
        return total_protect_rids

    def check_pf_feasibility(self):
        """check that current model still satisfies the protected functions
        """
        feasible = True
        for pf in self.nrp.protected_functions.values():
            self.set_temp_func_conditions(pf)
            res = self.fba_model.fba_optimize(short_result=True)
            self.restore_base_conditions()
            if res['success'] is False:
                feasible = False
                break
            if pf.obj_dir == 'maximize':
                if res['fun'] < pf.target_lb:
                    feasible = False
                    break
            else:
                if res['fun'] > pf.target_ub:
                    feasible = False
                    break
        return feasible

    def set_temp_func_conditions(self, pf):
        self.base_obj_dir = self.fba_model.obj_dir
        self.base_flux_bounds = self.fba_model.flux_bounds.copy()
        self.base_coefs = self.fba_model.obj_coefs.copy()

        self.fba_model.obj_dir = pf.obj_dir
        self.fba_model.update_flux_bounds(pf.overwrite_bounds)
        self.fba_model.obj_coefs = self.fba_model.set_objective_coefficients(pf.objective)

    def restore_base_conditions(self):
        self.fba_model.obj_dir = self.base_obj_dir
        self.fba_model.flux_bounds = self.base_flux_bounds
        self.fba_model.obj_coefs = self.base_coefs

    def get_reduction_status(self):
        """ Return status in model reduction process

        degrees of freedom, number removed reactions, number essential reactions
        number remaining reactions
        :returns: status information in a dictionary
        :rtype: dict
        """
        status = self.fba_model.get_model_status()
        status['n_protected_rids'] = len(self.protected_rids)
        status['n_protected_sids'] = len(self.protected_sids)
        status['essential_rids'] = len(self.essential_rids)
        return status

    def print_status(self):

        status = self.get_reduction_status()
        print("reactions removed/remaining/protected: "
              f"{status['n_full_rids' ] -status['n_red_rids']:4d}/{status['n_red_rids']:4d}/"
              f"{status['essential_rids']:4d}; dof: {status['dof']:3d}")
