

import math
import numpy as np
import sbmlxdf

from modelpruner.problems.lp_problem import LpProblem


class FbaModel:

    def __init__(self, sbml_model):

        model_dict = sbml_model.to_df()
        df_reactions = model_dict['reactions']
        df_species = model_dict['species']
        df_fbc_objectives = model_dict['fbcObjectives']

        self.full_model_shape = {'n_full_rids': len(df_reactions),
                                 'n_full_sids': len(df_species)}
        self.rids = df_reactions.index.to_numpy()
        self.rid2idx = {rid: idx for idx, rid in enumerate(self.rids)}
        self.sids = df_species.index.to_numpy()
        self.sid2idx = {sid: idx for idx, sid in enumerate(self.sids)}
        self.reversible = df_reactions['reversible'].to_numpy().copy()
        self.flux_bounds = np.vstack((df_reactions['fbcLb'].to_numpy(),
                                      df_reactions['fbcUb'].to_numpy()))
        self.s_mat = sbml_model.get_s_matrix().to_numpy()

        self.objective = {}
        for oid, row in df_fbc_objectives.iterrows():
            if row['active'] is True:
                self.obj_id = row.name
                self.obj_dir = row['type']
                for record in sbmlxdf.misc.record_generator(row['fluxObjectives']):
                    params = sbmlxdf.misc.extract_params(record)
                    self.objective[params['reac']] = params['coef']

        self.check_consistency()
        self.obj_coefs = self.set_objective_coefficients(self.objective)

    def set_objective_coefficients(self, objective):
        obj_coefs = np.zeros(len(self.rids))
        for rid, coef in objective.items():
            obj_coefs[self.rid2idx[rid]] = coef
        return obj_coefs

    def check_consistency(self):

        # check that irreversible reaction have a flux lower bound >= 0

        for idx in range(len(self.rids)):
            if np.logical_not(self.reversible[idx]):
                if self.flux_bounds[0, idx] < 0:
                    self.reversible[idx] = True
                    print(f'reaction {self.rids[idx]} set to reversible')

        if len(self.objective) == 0:
            print('Active FBA objective function missing in the model!')

        for rid in self.objective:
            if rid not in self.rids:
                print('objective reaction not part of reactions')
                raise AttributeError

    def combine_fwd_bwd(self, vector):
        """combine forward and backward flux information.

        :param vector: contains data on all reactions + backward information for reversible reactions
        :type vector: 1D ndarray of shape (len(mr_rids) + sum(reversible))
        :return: combined flux information
        :rtype: 1D ndarray of shape (len(mr_rids))
        """
        n_cols = len(self.rids)
        combined = vector[:n_cols]
        combined[self.reversible] -= vector[n_cols:]
        return combined

    def create_fba_lp(self):
        """create Linear Programming Problem for FBA with split reactions

        Splitting reactions in forward/backward directions reduces high values during FBA
        Splitting required for pFBA (minimize absolute flux values)

        :return: LpProblem configured for FBA
        :rtype: LpProblem
        """
        lp = LpProblem()
        fwd_bwd_s_mat = np.hstack((self.s_mat, -self.s_mat[:, self.reversible]))
        fwd_bwd_obj_coefs = np.hstack((self.obj_coefs, -self.obj_coefs[self.reversible]))
        fwd_bwd_bnds = np.hstack((np.fmax(self.flux_bounds, 0),
                                  np.vstack((np.fmax(-self.flux_bounds[1, self.reversible], 0),
                                             np.fmax(-self.flux_bounds[0, self.reversible], 0)))))
        row_bnds = np.zeros((fwd_bwd_s_mat.shape[0], 2))
        lp.configure(fwd_bwd_obj_coefs, fwd_bwd_s_mat, row_bnds, fwd_bwd_bnds, self.obj_dir)
        return lp

    def fba_optimize(self, short_result=False):
        """FBA optimization of current problem

        :return: results information from the optimization
        :rtype: dict
        """
        lp = self.create_fba_lp()
        if lp is None:
            return {'success': False, 'message': 'not an FBA model'}
        res = lp.solve(short_result)
        if short_result is False:
            res['x'] = self.combine_fwd_bwd(res['x'])
            res['reduced_costs'] = self.combine_fwd_bwd(res['reduced_costs'])
        return res

    def fva_optimize(self, rids=None, fract_of_optimum=1.0):
        """FVA - flux variability analysis for all or selected reactions.

        :param rids: selected reaction ids to run FVA (alternatively, all)
        :type rids: list of strings (default: None - all reactions)
        :param fract_of_optimum: value between 0.0 and 1.0.
        :type fract_of_optimum: float (default 1.0)
        :return: rid, min and max flux values
        :rtype: dict
        """
        lp = self.create_fba_lp()
        if lp is None:
            return {'success': False, 'message': 'not an FBA model'}
        res = lp.solve(short_result=True)

        if rids is None:
            rids = self.rids

        flux_min = np.zeros(len(rids))
        flux_max = np.zeros(len(rids))
        if res['success']:
            # fix FBA objective as constraint
            fwd_bwd_obj_coefs = np.hstack((self.obj_coefs, -self.obj_coefs[self.reversible]))
            if self.obj_dir == 'maximize':
                row_bds = np.array([res['fun'] * fract_of_optimum, np.nan])
            else:
                if fract_of_optimum > 0.0:
                    row_bds = np.array([np.nan, res['fun'] / fract_of_optimum])
                else:
                    row_bds = np.array([np.nan, 1000.0])
            lp.add_constraint(fwd_bwd_obj_coefs, row_bds)

            n_cols = len(self.rids)
            for idx, rid in enumerate(rids):
                # select reaction maximize/minimize flux
                new_obj_coefs = np.zeros(n_cols)
                new_obj_coefs[self.rid2idx[rid]] = 1
                new_fwd_bwd_obj_coefs = np.hstack((new_obj_coefs, -new_obj_coefs[self.reversible]))
                lp.set_obj_coefs(new_fwd_bwd_obj_coefs)

                lp.set_obj_dir('minimize')
                res = lp.solve(short_result=True)
                flux_min[idx] = res['fun'] if res['success'] else np.nan

                lp.set_obj_dir('maximize')
                res = lp.solve(short_result=True)
                flux_max[idx] = res['fun'] if res['success'] else np.nan

        return {'rids': np.array(rids), 'min': flux_min, 'max': flux_max}

    def update_flux_bounds(self, flux_bounds):
        """selectivly set lower/upper flux bounds for given reactions.

        Flux bounds with value of 'np.nan' will not be set

        :param flux_bounds: new lower and upper flux bounds selected reactions
        :type flux_bounds: dict (key: reaction id, value: list with two float values for lower/upper bound)
        """
        for rid, (lb, ub) in flux_bounds.items():
            if math.isfinite(lb):
                self.flux_bounds[0, self.rid2idx[rid]] = lb
                if lb < 0:
                    self.reversible[self.rid2idx[rid]] = True
            if math.isfinite(ub):
                self.flux_bounds[1, self.rid2idx[rid]] = ub

    def remove_reactions(self, drop_rids, protected_sids):
        # remove_reactions
        # also remove corresponding metabolite columns
        #  update, rids, rid2idx, sids, sid2idx, reversible, flux_bounds, s_mat

        keep_rids_mask = np.array([False if rid in drop_rids else True for rid in self.rids])

        self.rids = self.rids[keep_rids_mask]
        self.rid2idx = {rid: idx for idx, rid in enumerate(self.rids)}
        self.reversible = self.reversible[keep_rids_mask]
        self.obj_coefs = self.obj_coefs[keep_rids_mask]
        self.flux_bounds = self.flux_bounds[:, keep_rids_mask]
        self.s_mat = self.s_mat[:, keep_rids_mask]

        keep_sids_mask = np.any(self.s_mat != 0, axis=1)
        drop_sids = set(self.sids[np.logical_not(keep_sids_mask)])

        if len(drop_sids) > 0:
            drop_protected_sids = drop_sids.intersection(protected_sids)
            if len(drop_protected_sids) > 0:
                print('Something went worong, protected metabolites dropped due to reaction removal.',
                      drop_protected_sids)
            self.sids = self.sids[keep_sids_mask]
            self.sid2idx = {sid: idx for idx, sid in enumerate(self.sids)}
            self.s_mat = self.s_mat[keep_sids_mask, :]

    def get_identical_columns(self):
        # Identify identical columns:
        # sort the columns
        # check parallel columns, if they are identical
        # report on identical columns

        # we also consider reversibility
        # we do not consider linearly scaled reactions
        fwd_bwd_s_mat = np.hstack((self.s_mat, -self.s_mat[:, self.reversible]))
        fwd_bwd_rids = np.hstack((self.rids, self.rids[self.reversible]))

        # identify groups of parallel reactions
        sorted_idx = np.lexsort(fwd_bwd_s_mat)
        groups = []
        consumed_idxs = set()
        for pos in range(len(sorted_idx)):
            idx = sorted_idx[pos]
            if idx not in consumed_idxs:
                group = [idx]
                for adjacent_pos in range(pos + 1, len(sorted_idx)):
                    candiate_idx = sorted_idx[adjacent_pos]
                    if np.all(fwd_bwd_s_mat[:, idx] == fwd_bwd_s_mat[:, candiate_idx]):
                        group.append(candiate_idx)
                        consumed_idxs.add(candiate_idx)
                    else:
                        break
                if len(group) > 1:
                    groups.append(group)

        parallel_rid_groups = [list(fwd_bwd_rids[group]) for group in groups]
        return parallel_rid_groups

    def get_model_status(self):
        """ Return status on model


        :returns: status information in a dictionary
        :rtype: dict
        """
        status = {'dof': self.s_mat.shape[1] - np.linalg.matrix_rank(self.s_mat),
                  'n_red_rids': len(self.rids),
                  'n_red_sids': len(self.sids)}
        status.update(self.full_model_shape)
        return status

    @property
    def obj_dir(self):
        return self.__obj_dir

    @obj_dir.setter
    def obj_dir(self, dir_string):
        if 'min' in dir_string:
            self.__obj_dir = 'minimize'
        else:
            self.__obj_dir = 'maximize'
